//-----------------------------------------------------------------------------
//
//  DRAW2D
//
//-----------------------------------------------------------------------------
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <vector>

//d3d
#include <d3d9.h>
#include <d3dx9.h>


#define RELEASE(x)  { if(x) (x)->Release(); x = NULL; }

//IS TEST?
#define _DRAW2D_TEST_

//pragma
#pragma comment(lib, "winmm.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "gdi32.lib")
#pragma comment(lib, "dxguid.lib")
#pragma comment(lib, "d3d9.lib")
#pragma comment(lib, "d3dx9.lib")
#pragma comment(lib, "dinput8.lib")
#pragma comment(lib, "dsound.lib")


//DirectGraphics
LPDIRECT3D9           D3D9Object = NULL;
LPDIRECT3DDEVICE9     D3D9Device = NULL;
static D3DPRESENT_PARAMETERS d3dpp  = {
  640, 480, D3DFMT_X8R8G8B8, 1, D3DMULTISAMPLE_NONE, 0,
  D3DSWAPEFFECT_DISCARD, 0, TRUE, TRUE, D3DFMT_D16, 0, 0, TRUE,
};

//-----------------------------------------------------------------------------
//
// TEXTURE
//
//-----------------------------------------------------------------------------
#define TEXTURE_MAX    256
struct TEXTUREINFO {
  LPDIRECT3DTEXTURE9 texture;
  char szName[MAX_PATH];
};
static TEXTUREINFO textureinfo[TEXTURE_MAX];

//-----------------------------------------------------------------------------
// TextureOpen
//-----------------------------------------------------------------------------
HRESULT TextureClose(UINT index) {
  if(index < TEXTURE_MAX) {
    TEXTUREINFO *w = &(textureinfo[index]);
    RELEASE(w->texture);
    ZeroMemory(w->szName, sizeof(w->szName));
  }
  return S_OK;
}

//-----------------------------------------------------------------------------
// TextureOpen
//-----------------------------------------------------------------------------
HRESULT TextureOpen(UINT index, LPSTR szName) {
  HRESULT             hRet  = E_FAIL;
  LPDIRECT3DTEXTURE9  lpTex = NULL;
  D3DXIMAGE_INFO      iinfo;
  if(D3D9Device && index < TEXTURE_MAX) {
    TEXTUREINFO *w = &(textureinfo[index]);
    TextureClose(index);
    if( D3DXGetImageInfoFromFile(szName, &iinfo) == D3D_OK) {
      hRet = D3DXCreateTextureFromFileEx(
        D3D9Device, szName,
        iinfo.Width, iinfo.Height,
        1, 0,
        D3DFMT_A8R8G8B8,
        D3DPOOL_MANAGED,
        D3DX_FILTER_NONE, D3DX_FILTER_NONE,
        0xFF000000, NULL, NULL,
        &w->texture);
      if(hRet == S_OK) {
        wsprintf(w->szName, "%s", szName);
      }
    }
  }
  return hRet;
}

//-----------------------------------------------------------------------------
// TextureOnLost
//-----------------------------------------------------------------------------
HRESULT TextureOnLost() {
  for(int i = 0 ; i < TEXTURE_MAX; i++) {
    //Only Texture container.
    TEXTUREINFO *w = &(textureinfo[i]);
    RELEASE(w->texture);
  }
  return S_OK;
}

//-----------------------------------------------------------------------------
// TextureRestore
//-----------------------------------------------------------------------------
HRESULT TextureRestore() {
  HRESULT hRet = S_OK;
  char szName[MAX_PATH];
  for(int i = 0 ; i < TEXTURE_MAX; i++) {
    TEXTUREINFO *w = &(textureinfo[i]);
    wsprintf(szName, "%s", w->szName);
    if(strlen(szName) > 0) {
      hRet |= TextureOpen(i, szName);
    }
  }
  return hRet;
}

//-----------------------------------------------------------------------------
// TextureSet
//-----------------------------------------------------------------------------
HRESULT TextureSet(int index) {
  HRESULT             hRet = E_FAIL;
  if(D3D9Device && index < TEXTURE_MAX) {
    TEXTUREINFO *w = &(textureinfo[index]);
    hRet = D3D9Device->SetTexture(0, w->texture);
  }
  return hRet;
}



//-----------------------------------------------------------------------------
// DeviceOnLost
//-----------------------------------------------------------------------------
static HRESULT DeviceOnLost() {
  TextureOnLost();
  return S_OK;
}

//-----------------------------------------------------------------------------
// DeviceOnRestore
//-----------------------------------------------------------------------------
static HRESULT DeviceOnRestore() {
  if(D3D9Device) {
    TextureRestore();
    D3D9Device->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
    D3D9Device->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
    D3D9Device->SetRenderState( D3DRS_ZENABLE, TRUE );
    D3D9Device->SetRenderState( D3DRS_AMBIENT, 0xffffffff );
  }
  return S_OK;
}


//-----------------------------------------------------------------------------
// DeviceBeginScene
//-----------------------------------------------------------------------------
HRESULT DeviceBeginScene() {
  HRESULT hr = E_FAIL;
  if(D3D9Device) {
    hr = D3D9Device->TestCooperativeLevel();
    if(hr == D3DERR_DEVICENOTRESET) {
      if(DeviceOnLost() != S_OK) {
        return hr;
      }
      if(D3D9Device->Reset(&d3dpp) != D3D_OK) {
        return hr;
      }
      if(DeviceOnRestore() != S_OK) {
        return hr;
      }
    }
    hr = D3D9Device->BeginScene();
  }
  return hr;
}

//-----------------------------------------------------------------------------
// DeviceClearScene
//-----------------------------------------------------------------------------
HRESULT DeviceClearScene(D3DCOLOR col) {
  HRESULT hRet = E_FAIL;
  if(D3D9Device) {
    hRet = D3D9Device->Clear(NULL, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, col, 1.0f, 0);
  }
  return hRet;
}

//-----------------------------------------------------------------------------
// DeviceEndScene
//-----------------------------------------------------------------------------
HRESULT DeviceEndScene() {
  HRESULT hRet = E_FAIL;
  if(D3D9Device) {
    hRet = D3D9Device->EndScene();
  }
  return hRet;
}

//-----------------------------------------------------------------------------
// DevicePresentScene
//-----------------------------------------------------------------------------
HRESULT DevicePresentScene() {
  HRESULT hRet = E_FAIL;
  if(D3D9Device) {
    hRet = D3D9Device->Present(0, 0, 0, 0);
  }
  return hRet;
}



//-----------------------------------------------------------------------------
// STRUCT
//-----------------------------------------------------------------------------
struct D4_DIFF_TEX {
  D3DXVECTOR4 p;
  DWORD color;
  FLOAT u, v;
  enum { FVF = D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1 };
};


struct DRAW2DINFO {
  D4_DIFF_TEX v[6];
};
static std::vector<DRAW2DINFO> vDraw2D;


HRESULT Draw2DPush(
  FLOAT x1, FLOAT y1,
  FLOAT x2, FLOAT y2,
  D3DCOLOR color,
  FLOAT u1, FLOAT v1, FLOAT u2, FLOAT v2)
{
  HRESULT hRet = E_FAIL;
  if(D3D9Device) {
    x1 -= 0.5f; y1 -= 0.5f; x2 -= 0.5f; y2 -= 0.5f;
    DRAW2DINFO temp = {
      { { D3DXVECTOR4(x1, y1, 0.0f, 1.0f), color, u1, v1},
        { D3DXVECTOR4(x2, y1, 0.0f, 1.0f), color, u2, v1},
        { D3DXVECTOR4(x1, y2, 0.0f, 1.0f), color, u1, v2},
        { D3DXVECTOR4(x2, y1, 0.0f, 1.0f), color, u2, v1},
        { D3DXVECTOR4(x2, y2, 0.0f, 1.0f), color, u2, v2},
        { D3DXVECTOR4(x1, y2, 0.0f, 1.0f), color, u1, v2}, }
    };
    vDraw2D.push_back(temp);
  }
  return hRet;
}

HRESULT Draw2DRotatePush(
  FLOAT x, FLOAT y,
  FLOAT w, FLOAT h,
  D3DCOLOR color,
  FLOAT u1, FLOAT v1, FLOAT u2, FLOAT v2, FLOAT angle)
{
  HRESULT hRet = E_FAIL;
  if(D3D9Device) {
    D3DXMATRIX  mRot;
    D3DXMATRIX  mTrans;
    D3DXVECTOR4 v[4];
    x -= 0.5f; y -= 0.5f;// w -= 0.5f; h -= 0.5f;
    D3DXVECTOR2 t[4] = {
      D3DXVECTOR2(- w, - h), D3DXVECTOR2(+ w, - h), D3DXVECTOR2(- w, + h), D3DXVECTOR2(+ w, + h),
    };
    D3DXMatrixTranslation(&mTrans, x, y, 0);
    D3DXMatrixRotationZ(&mRot, angle);
    for(int i = 0 ; i < 4; i++) D3DXVec2Transform(&(v[i]), &(t[i]), &mRot);
    for(int i = 0 ; i < 4; i++) D3DXVec4Transform(&(v[i]), &(v[i]), &mTrans);
    DRAW2DINFO temp =
    {
      { { v[0], color, u1, v1}, { v[1], color, u2, v1}, { v[2], color, u1, v2},
        { v[1], color, u2, v1}, { v[3], color, u2, v2}, { v[2], color, u1, v2}, }
    };
    vDraw2D.push_back(temp);
  }
  return hRet;
}

HRESULT Draw2DPresent() {
  HRESULT hRet = E_FAIL;
  if(D3D9Device) {
    if(!vDraw2D.empty()) {
      hRet = D3D9Device->SetFVF(D4_DIFF_TEX::FVF);
      hRet = D3D9Device->DrawPrimitiveUP(D3DPT_TRIANGLELIST, vDraw2D.size() * 2, &(vDraw2D[0]), sizeof(D4_DIFF_TEX));
    }
  }
  vDraw2D.clear();
  return hRet;
}


HRESULT DeviceTerm() {
  for(int i = 0 ; i < TEXTURE_MAX; i++) {
    TextureClose(i);
  }
  RELEASE(D3D9Device);
  RELEASE(D3D9Object);
  return S_OK;
}

HRESULT DeviceInit(HWND hWnd, UINT Width, UINT Height, BOOL Windowed) {
  HRESULT hRet = E_FAIL;
  if(D3D9Object) return hRet;
  ZeroMemory(textureinfo, sizeof(textureinfo));
  
  //Setup param
  d3dpp.BackBufferWidth = Width;
  d3dpp.BackBufferHeight = Height;
  d3dpp.Windowed = Windowed;
  
  //Create
  D3D9Object = Direct3DCreate9(D3D_SDK_VERSION);
  if(!D3D9Object) {
    DeviceTerm();
    return hRet;
  }
  
  hRet = D3D9Object->CreateDevice(
    D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &D3D9Device);
  if(!D3D9Device) {
    DeviceTerm();
    return hRet;
  }
  DeviceOnRestore();
  return hRet;
}

//-----------------------------------------------------------------------------
//
//
//  _DRAW2D_TEST_
//
//
//-----------------------------------------------------------------------------

#ifdef _DRAW2D_TEST_

#define WIDTH    1024
#define HEIGHT   768

//-----------------------------------------------------------------------------
//  Function
//-----------------------------------------------------------------------------
static LRESULT WINAPI AppWinProc(HWND hwindow, UINT msg, WPARAM wParam, LPARAM lParam);

//-----------------------------------------------------------------------------
//  Var
//-----------------------------------------------------------------------------
static char  AppName[MAX_PATH];
static HWND  hWnd = NULL;

//Window Class Ex
static WNDCLASSEX wcex = {
  sizeof(WNDCLASSEX), CS_CLASSDC | CS_VREDRAW | CS_HREDRAW,
  AppWinProc, 0L, 0L, NULL, ::LoadIcon(NULL, IDI_APPLICATION),
  ::LoadCursor(NULL, IDC_ARROW), (HBRUSH)::GetStockObject(BLACK_BRUSH),
  NULL, AppName, NULL
};


//-----------------------------------------------------------------------------
//  Window Proc
//-----------------------------------------------------------------------------
static LRESULT WINAPI AppWinProc(HWND hwindow, UINT msg, WPARAM wParam, LPARAM lParam) {
  switch(msg) {
    case WM_SYSCOMMAND:
      switch((wParam & 0xFFF0)) {
        case SC_MONITORPOWER:
        case SC_SCREENSAVE:
          return 0;
          break;
      }
      break;
    /*
    case WM_ACTIVATE:
      if(wParam == WA_INACTIVE) bActive = FALSE;
      else                      bActive = TRUE;
      break;
    */
    case WM_IME_SETCONTEXT:
      lParam &= ~ISC_SHOWUIALL;
      break;
    case WM_CLOSE:
    case WM_DESTROY:
      ::PostQuitMessage(0);
      return 0;
      break;
    case WM_KEYDOWN:
      if(wParam == VK_ESCAPE) ::PostQuitMessage(0);
      break;
  }
  return ::DefWindowProc(hwindow, msg, wParam, lParam);
}


//-----------------------------------------------------------------------------
//  AppDo Event
//-----------------------------------------------------------------------------
BOOL AppDoEvent() {
  MSG  msg;
  BOOL bActive = TRUE;
  while(::PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
    if(msg.message == WM_QUIT) {
      bActive = FALSE;
      break;
    } else {
      ::TranslateMessage(&msg);
      ::DispatchMessage(&msg);
    }
  }
  return bActive;
}

//-----------------------------------------------------------------------------
//  Quit
//-----------------------------------------------------------------------------
void AppQuit() {
  PostQuitMessage(0);
}

//-----------------------------------------------------------------------------
//  StartApp
//-----------------------------------------------------------------------------
HRESULT AppInit(LPSTR szName, UINT Width, UINT Height, BOOL Windowed) {
  HRESULT hRet = E_FAIL;
  DWORD exstyle  = WS_EX_APPWINDOW | Windowed ? WS_EX_WINDOWEDGE : 0;
  DWORD style    = Windowed  ? WS_OVERLAPPEDWINDOW : WS_POPUP;
  if(hWnd) return E_FAIL;
  wsprintf(AppName, "%s", szName);

  //Create Instance.
  wcex.hInstance = GetModuleHandle(NULL);
  if(!::RegisterClassEx(&wcex)) return hRet;
  hWnd = ::CreateWindowEx(exstyle, AppName, AppName, style, 0, 0, Width, Height, NULL, NULL, wcex.hInstance, NULL);

  //Display Window
  ::ShowWindow(hWnd, SW_SHOW);
  ::UpdateWindow(hWnd);
  Windowed ? (void)0 : ShowCursor(FALSE);
  
  return hRet = S_OK;
}

int main() {
  do {
    if(AppInit("DRAW2DTEST", WIDTH, HEIGHT, TRUE) != S_OK) break;
    if(DeviceInit(hWnd, WIDTH, HEIGHT, TRUE) != S_OK)      break;
    
    //Load
    TextureOpen(0, "va.jpg");
    while(AppDoEvent()) {
      static FLOAT angle = 0;
      angle += 0.01;
      for(int i = 0 ; i < 2560; i++) {
        Draw2DRotatePush(rand() % WIDTH, rand() % HEIGHT, 32, 32, (rand() << 16) | rand(),
          0, 0, 1, 1,
          angle);
      }
      DeviceClearScene(0xFF00FF00);
      if(DeviceBeginScene() == S_OK) {
        TextureSet(0);
        Draw2DPresent();
        
        //Draw
        Draw2DPush(0, 0, WIDTH, HEIGHT, 0xFFFFFFFF, 0, 0, 1, 1);
        Draw2DPresent();
        DeviceEndScene();
      }
      DevicePresentScene();
    }
  } while(0);
  DeviceTerm();
  return 0;
}

#endif // _DRAW2D_TEST_

